// ------------------------------------------------------------------------
// Filter Project -- Tecgraf/PUC-Rio
// www.tecgraf.puc-rio.br
//
// $Source: $
// $Revision: $
// $Date: 2007/05/1 10:07:00 $
//
// $Author: lduarte $
// ------------------------------------------------------------------------

#include <math.h>
#include <stdlib.h>
#include <assert.h>

#include "filter.h"

static int resolution = 1024;
static float* WTAB;

int Min(int a, int b) {return a < b ? a : b;}
int Max(int a, int b) {return a > b ? a : b;}

// ------------------------------------------------------------------------
// ------------------------------------------------------------------------
void initWTAB(void)
{
 int i;
 float r, alpha = 2;
 WTAB = (float*)malloc(resolution*sizeof(float));
 for(i = 0; i < resolution; i++)
 {
  r = (float)i/(float)(resolution - 1);
  WTAB[i] = (float)exp(-alpha*r);
 }
}

// ------------------------------------------------------------------------
// ------------------------------------------------------------------------
void EWAFilter(Image* texture, double u0, double v0, double ux, double uy, double vx, double vy, float* rgb)
{
 /* Ellipse coefficients */
 float texel[3], NUM[3];
 int u1, u2, v1, v2, u, v, index;
 double DEN, DDQ, DQ, r, U, V;
 double A, B, C, F, Q, WEIGHT, scale;

 A = vx*vx + vy*vy + 1;
 B = -2*(ux*vx + uy*vy);
 C = ux*ux + uy*uy + 1;
 F = A*C - B*B/4;

 scale = resolution/F;
 A *= scale;
 B *= scale;
 C *= scale;
 F = resolution;

 /* Ellipse bounding box */
	u1 = (int)floor(u0 - (2.0/(( -B*B + 4.0*C*A)*sqrt((-B*B + 4.0*C*A)*C*F))));
	u2 = (int)ceil(u0 + (2.0/(( -B*B + 4.0*C*A)*sqrt((-B*B + 4.0*C*A)*C*F))));
	v1 = (int)floor(v0 - (2.0/((-B*B + 4.0*C*A)*sqrt((-B*B + 4.0*C*A)*A*F))));
	v2 = (int)ceil(v0 + (2.0/((-B*B + 4.0*C*A)*sqrt((-B*B + 4.0*C*A)*A*F))));

 u1 = Max(u1, 0 );
	u2 = Min(u2, imgGetWidth(texture)-1);
	v1 = Max(v1, 0 );
	v2 = Min(v2, imgGetHeight(texture)-1);

 NUM[0] = 0.0;
 NUM[1] = 0.0;
 NUM[2] = 0.0;

 DEN = 0.0;
	DDQ = 2.0*A;
	U = u1 - u0;
	for(v = v1; v <= v2; v++)
	{
		V = v - v0;
		DQ = A*(2*U + 1)+ B*V;
		Q = (C*V + B*U)*V + A*U*U;
		for(u = u1; u <= u2; u++)
		{
			if(Q < F)
			{
				r = (Q/F);
    index = (int)(r*(resolution-1));
    WEIGHT = WTAB[index];

				imgGetPixel3fv(texture, u, v, texel);
				NUM[0] += WEIGHT*texel[0];
				NUM[1] += WEIGHT*texel[1];
				NUM[2] += WEIGHT*texel[2];
				
    DEN += WEIGHT;
			}
			Q += DQ;
			DQ += DDQ;
		}
	}

	if(DEN > 0.0)
 {
  rgb[0] = (float)NUM[0]/DEN;
	 rgb[1] = (float)NUM[1]/DEN;
	 rgb[2] = (float)NUM[2]/DEN;
 }
}

// ------------------------------------------------------------------------
// ------------------------------------------------------------------------
void EccentricityCorr(double *du, double *dv)
{
 double lu, lv, eccent;
 // Compare lengths
	lu = sqrt((du[0]*du[0]) + (du[1]*du[1]));
	lv = sqrt((dv[0]*dv[0]) + (dv[1]*dv[1]));

	if(lv < lu)
	{
		eccent = lu/lv;
  if(eccent > 5.0)
  {
   eccent /= 5.0;
   dv[0] *= eccent;
   dv[1] *= eccent;
  }
	}
	else if(lu < lv)
	{
		eccent = lv/lu;
  if(eccent > 5.0)
  {
   eccent /= 5.0;
   du[0] *= eccent;
   du[1] *= eccent;
  }
	}
}

// ------------------------------------------------------------------------
// ------------------------------------------------------------------------
/* Calcula a variaao do mapeamento (x,y) nos seus vizinhos */
void derivateMap(double* mapInv, int x, int y, double* du, double* dv)
{
	/* vetor derivada na direcao u */
 du[0] = ((mapInv[0]*mapInv[7]-mapInv[1]*mapInv[6])*y+mapInv[0]*mapInv[8]-mapInv[6]*mapInv[2]);
	du[1] = ((mapInv[1]*mapInv[6]-mapInv[0]*mapInv[7])*x+mapInv[8]*mapInv[1]-mapInv[2]*mapInv[7]);

 /* dividindo pela coordenada homegenea */
 du[0] /= (mapInv[6]*x+mapInv[7]*y+mapInv[8])*(mapInv[6]*x+mapInv[7]*y+mapInv[8]);
 du[1] /= (mapInv[6]*x+mapInv[7]*y+mapInv[8])*(mapInv[6]*x+mapInv[7]*y+mapInv[8]);
	
 /* vetor derivada na direcao v */
 dv[0] = ((mapInv[3]*mapInv[7]-mapInv[4]*mapInv[6])*y+mapInv[3]*mapInv[8]-mapInv[5]*mapInv[6]);
	dv[1] = ((mapInv[4]*mapInv[6]-mapInv[3]*mapInv[7])*x+mapInv[4]*mapInv[8]-mapInv[5]*mapInv[7]);

 /* dividindo pela coordenada homegenea */
 dv[0] /= (mapInv[6]*x+mapInv[7]*y+mapInv[8])*(mapInv[6]*x+mapInv[7]*y+mapInv[8]);
 dv[1] /= (mapInv[6]*x+mapInv[7]*y+mapInv[8])*(mapInv[6]*x+mapInv[7]*y+mapInv[8]);
}


// ------------------------------------------------------------------------
// ------------------------------------------------------------------------
void BilinearFilter(Image* image, float *rgb, double u0, double v0)
{
	double deltaU, deltaV;
	float rgb1[3], rgb2[3], rgb3[3], rgb4[3];

	imgGetPixel3fv(image, (int)floor(u0), (int) ceil(v0), rgb1);
	imgGetPixel3fv(image, (int) ceil(u0), (int) ceil(v0), rgb2);
	imgGetPixel3fv(image, (int)floor(u0), (int)floor(v0), rgb3);
	imgGetPixel3fv(image, (int) ceil(u0), (int)floor(v0), rgb4);

 deltaU = u0 - (int)u0;
	deltaV = v0 - (int)v0;

 rgb[0] = (float)(((1-deltaU)*deltaV*rgb1[0])+(deltaU*deltaV*rgb2[0])+
          ((1-deltaU)*(1-deltaV)*rgb3[0])+(deltaU*(1-deltaV)*rgb4[0]));

 rgb[1] = (float)(((1-deltaU)*deltaV*rgb1[1])+(deltaU*deltaV*rgb2[1])+
          ((1-deltaU)*(1-deltaV)*rgb3[1])+(deltaU*(1-deltaV)*rgb4[1]));

 rgb[2] = (float)(((1-deltaU)*deltaV*rgb1[2])+(deltaU*deltaV*rgb2[2])+
          ((1-deltaU)*(1-deltaV)*rgb3[2])+(deltaU*(1-deltaV)*rgb4[2]));
}


// ------------------------------------------------------------------------
// ------------------------------------------------------------------------
void freeMem(void)
{
 if(WTAB)
  free(WTAB);
}